package minischeme

import scala.util.parsing.input.Position
import scala.collection.immutable.ListMap
import SymbolicTreeModule._

/**
 * Interpreter for minischeme
 *
 * @author Iulian Dragos, Michel Schinz
 */
object Interpreter extends (List[Tree] => Unit) {
  // Values
  abstract class Value {
    override def toString(): String = this match {
      case Closure(_) => "<lambda>"
      case Vector(_) => "<vector>"
      case Integer(v) => v.toString()
    }
  }
  case class Closure(body: List[Value] => Value) extends Value
  case class Vector(values: Array[Value]) extends Value
  case class Integer(v: Int) extends Value

  // Error handling
  case class InterpreterError() extends Exception()

  def printStackEntry(pos: Position): Unit =
    Console.println("  " + pos)

  def error(pos: Option[Position], msg: String): Nothing = {
    Console.println("Error: " + msg)
    pos foreach printStackEntry
    throw new InterpreterError()
  }
  def error(pos: Position, msg: String): Nothing = error(Some(pos), msg)
  def error(msg: String): Nothing = error(None, msg)

  // Environment
  type Env = PartialFunction[Symbol, Value]

  val globalEnv = new scala.collection.mutable.HashMap[Symbol, Value]

  def apply(trees: List[Tree]): Unit = try {
    trees foreach {
      case Define(symbol, _) =>
        globalEnv += Pair(symbol, Integer(0))
      case _ =>
    }
    trees foreach {
      case Define(symbol, value) =>
        globalEnv += Pair(symbol, eval(value, globalEnv))
      case tree =>
        eval(tree, globalEnv)
    }
  } catch {
    case InterpreterError() => ()
  }

  def eval(tree: Tree, env: Env): Value = {
    def t_error(msg: String): Nothing = error(Some(tree.pos), msg)

    tree match {
      case Lambda(params, body) =>
	Closure({ args: List[Value] =>
	  if (args.length == params.length)
	    eval(body, ListMap(params zip args : _*) orElse env)
	  else
	    error("wrong number of arguments (expected " + params.length
                  + ", got " + args.length + ")")
        })

      case Let(ident, value, body) =>
	eval(body, ListMap((ident, eval(value, env))) orElse env)

      case Begin(body) =>
        body.foldLeft(null : Value) { (_, t) => eval(t, env) }

      case If(cond, thenp, elsep) =>
	eval(cond, env) match {
	  case Integer(0) => eval(elsep, env)
	  case _ => eval(thenp, env)
	}

      case And(c1, c2) =>
        eval(c1, env) match {
          case res @ Integer(0) => res
          case _ => eval(c2, env)
        }

      case Or(c1, c2) =>
        eval(c1, env) match {
          case Integer(0) => eval(c2, env)
          case res => res
        }

      case Apply(fun, args) =>
        eval(fun, env) match {
	  case Closure(body) =>
            try {
              body(args map { a: Tree => eval(a, env) })
            } catch {
              case e @ InterpreterError() =>
                printStackEntry(tree.pos)
                throw e
            }
	  case v =>
            t_error("expected function, got " + v)
	}

      case Prim(Primitives.read_char, List()) =>
        Integer(System.in.read())

      case Prim(Primitives.print_char, List(c)) =>
        eval(c, env) match {
          case Integer(i) =>
            System.out.print(i.asInstanceOf[Char])
            System.out.flush()
            Integer(0)
	  case other =>
            t_error("integer expected, got " + other)
	}

      case Prim(Primitives.vector, args) =>
        Vector((args map { a: Tree => eval(a, env) }).toArray)

      case Prim(Primitives.vector_ref, List(vect, index)) =>
        (eval(vect, env), eval(index, env)) match {
          case (Vector(contents), Integer(i)) =>
            if (0 <= i && i < contents.length)
              contents(i)
            else
              t_error("index " + i + " not in [0;" + contents.length + ")")
          case (other, Integer(_)) =>
            t_error("vector expected for vector-ref, got " + other)
          case (_, other) =>
            t_error("integer expected for vector-ref index, got " + other)
        }

      case Prim(Primitives.vector_set, List(vect, index, value)) =>
        (eval(vect, env), eval(index, env), eval(value, env)) match {
          case (Vector(contents), Integer(i), v) =>
            if (0 <= i && i < contents.length) {
              contents(i) = v
              Integer(0)
            } else
              t_error("index " + i + " not in [0;" + contents.length + ")")
          case (other, Integer(_), _) =>
            t_error("vector expected for vector-set!, got " + other)
          case (_, other, _) =>
            t_error("integer expected for vector-set! index, got " + other)
        }

      case Prim(sym, List(e1, e2)) =>
        implicit def intToInteger(i: Int): Integer = Integer(i)
        implicit def boolToInteger(b: Boolean): Integer = if (b) 1 else 0

        (sym, eval(e1, env), eval(e2, env)) match {
          case (Primitives.add, Integer(v1), Integer(v2)) => Integer(v1 + v2)
          case (Primitives.sub, Integer(v1), Integer(v2)) => Integer(v1 - v2)
          case (Primitives.mul, Integer(v1), Integer(v2)) => Integer(v1 * v2)
          case (Primitives.div, Integer(v1), Integer(v2)) => Integer(v1 / v2)
          case (Primitives.mod, Integer(v1), Integer(v2)) => Integer(v1 % v2)

          case (Primitives.lt, Integer(v1), Integer(v2)) => v1 < v2
          case (Primitives.le, Integer(v1), Integer(v2)) => v1 <= v2
          case (Primitives.eq, Integer(v1), Integer(v2)) => v1 == v2
          case (Primitives.ge, Integer(v1), Integer(v2)) => v1 >= v2
          case (Primitives.gt, Integer(v1), Integer(v2)) => v1 > v2

          case (Primitives.eq, Vector(v1), Vector(v2)) => v1 eq v2

          case (sym, v1, v2) =>
            t_error("invalid arguments for " + sym.name + ": " + v1 + " " + v2)
        }

      case Prim(Primitives.not, List(e)) =>
        eval(e, env) match {
          case Integer(i) => Integer(if (i == 0) 1 else 0)
          case other => t_error("integer expected, got " + other)
        }

      case Prim(sym, _) =>
        t_error("malformed use of primitive '" + sym + "'")

      case Ident(ident) =>
        env(ident)

      case Num(value)  =>
        Integer(value)
    }
  }
}
