package minischeme

import scala.util.parsing.input.Position
import scala.collection.immutable.{Set,Map}

import NominalTreeModule._
import SymbolicTreeModule.{Tree=>STree, Define=>SDefine, Lambda=>SLambda,
                           Let=>SLet, Begin=>SBegin, If=>SIf, And=>SAnd,Or=>SOr,
                           Apply=>SApply, Prim=>SPrim, Ident=>SIdent, Num=>SNum}

/**
 * Name analysis for minischeme: transforms a "nominal" tree, in which
 * identifiers are represented by names, to a "symbolic" tree, in
 * which identifiers are represented by symbols.
 *
 * @author Michel Schinz
 */
object NameAnalyzer extends ((List[Tree], Map[String,Symbol]) => List[STree]) {
  private type Env = Map[String,Symbol]

  def apply(trees: List[Tree], env: Env): List[STree] = {
    val globalEnv = collectGlobalEnv(trees, env)
    trees map { transform(_, globalEnv) }
  }

  private def collectGlobalEnv(trees: List[Tree], init: Env): Env =
    trees.foldLeft(init) { (env, tree) =>
      tree match {
        case Define(name, _) =>
          env + Pair(name, new Symbol(name, GlobalKind))
        case _ =>
          env
      }
    }

  private def transform(tree: Tree, env: Env): STree = copyPos(tree) {
    def checkDefs(syms: List[Symbol], defined: Set[String]): Unit =
      syms match {
        case n :: _ if (defined(n.name)) =>
          Report.error(tree.pos, "redefinition of " + n.name)
        case n :: ns => checkDefs(ns, defined + n.name)
        case _ => ()
      }

    tree match {
      case Define(name, value) =>
        SDefine(env(name), transform(value, env))
      case Lambda(params, body) =>
        val pSyms = params map { new Symbol(_, LocalKind) }
        checkDefs(pSyms, Set())
        SLambda(pSyms, transform(body, env ++ (pSyms map { s => (s.name, s)})))
      case Let(name, value, body) =>
        val sym = new Symbol(name, LocalKind)
        SLet(sym, transform(value, env), transform(body, env + ((name, sym))))
      case Begin(body) =>
        SBegin(transform(body, env))
      case If(c, t, e) =>
        SIf(transform(c, env), transform(t, env), transform(e, env))
      case And(c1, c2) =>
        SAnd(transform(c1, env), transform(c2, env))
      case Or(c1, c2) =>
        SOr(transform(c1, env), transform(c2, env))
      case Apply(fun, args) =>
        SApply(transform(fun, env), transform(args, env))
      case Prim(sym, args) =>
        SPrim(sym, transform(args, env))
      case Ident(name) if (env contains name) =>
        SIdent(env(name))
      case Ident(name) =>
        Report.error(tree.pos, "unknown identifier " + name)
      case Num(n) =>
        SNum(n)
    }
  }

  private def transform(trees: List[Tree], env: Env): List[STree] =
    trees map { transform(_, env) }

  private def copyPos(origTree: Tree)(newTree: STree): STree = {
    newTree.pos = origTree.pos
    newTree
  }
}
