package minischeme

import scala.util.parsing.combinator.syntactical.TokenParsers
import NominalTreeModule._

/**
 * Parser for minischeme.
 *
 * @author Michel Schinz
 */
class SExpParsers extends TokenParsers {
  type Tokens = SExpTokens
  val lexical = new SExpScanners

  def program: Parser[List[Tree]] =
    phrase(rep(define | expr))

  private def define: Parser[Tree] = positioned(
    lexical.LPar ~> lexical.Define ~> rIdent ~ expr <~ lexical.RPar
      ^^ { case i ~ es => Define(i, es) })
  private def expr: Parser[Tree] =
    pExpr | ident | num | str | chr
  private def pExpr: Parser[Tree] =
    lexical.LPar ~> (lambda|let|begin|cond|and|or|app) <~ lexical.RPar
  private def lambda: Parser[Tree] = positioned(
    lexical.Lambda ~> (lexical.LPar ~> rep(rIdent) <~ lexical.RPar) ~ rep1(expr)
      ^^ { case idents ~ exprs => Lambda(idents, desugarBodies(exprs)) })
  private def let: Parser[Tree] = positioned(
    lexical.Let ~> (lexical.LPar ~> rep(binding) <~ lexical.RPar) ~ rep1(expr)
      ^^ { case bdgs ~ expr => desugarLet(bdgs, expr)  })
  private def begin: Parser[Tree] = positioned(
    lexical.Begin ~> rep1(expr)
      ^^ { case exprs => Begin(exprs) })
  private def and: Parser[Tree] = positioned(
    lexical.And ~> expr ~ expr
      ^^ { case e1 ~ e2 => And(e1, e2) })
  private def or: Parser[Tree] = positioned(
    lexical.Or ~> expr ~ expr
      ^^ { case e1 ~ e2 => Or(e1, e2) })
  private def binding: Parser[(String,Tree)] =
    (lexical.LPar ~> rIdent ~ expr <~ lexical.RPar
      ^^ { case ident ~ expr => (ident, expr) })
  private def cond: Parser[Tree] = positioned(
    lexical.If ~> expr ~ expr ~ expr
      ^^ { case c ~ t ~ e => If(c, t, e) })
  private def app: Parser[Tree] = positioned(
    expr ~ rep(expr)
      ^^ { case f ~ as => Apply(f, as) })

  private def rIdent: Parser[String] = (
    elem("identifier", _.isInstanceOf[lexical.Ident])
      ^^ { case lexical.Ident(n) => n })
  private def ident: Parser[Ident] = positioned(rIdent ^^ Ident)
  private def num: Parser[Num] = positioned(
    elem("number", _.isInstanceOf[lexical.Num])
      ^^ { case lexical.Num(n) => Num(n) })
  private def str: Parser[Tree] = positioned(
    elem("string", _.isInstanceOf[lexical.Str])
      ^^ { case lexical.Str(s) => desugarStr(s) })
  private def chr: Parser[Tree] = positioned(
    elem("character", _.isInstanceOf[lexical.Chr])
      ^^ { case lexical.Chr(c) => Num(c.toInt) })

  private def desugarLet(bdgs: List[(String,Tree)], bodies: List[Tree]): Tree =
    (bdgs :\ desugarBodies(bodies)){ case ((n, v), b) => Let(n, v, b) }

  private def desugarBodies(bodies: List[Tree]): Tree = bodies match {
    case body :: Nil => body
    case bodies => Begin(bodies)
  }

  private def desugarStr(s: String): Tree =
    Prim(Primitives.vector,
         Num(s.length) :: s.foldRight(Nil : List[Tree])(Num(_) :: _))
}
