#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "engine.h"
#include "fail.h"

typedef struct _label_def_t {
  char* name;
  void* addr;
  struct _label_def_t* next;
} label_def_t;

typedef struct _label_ref_t {
  char* name;
  void* instr_addr;
  struct _label_ref_t* next;
} label_ref_t;

// labels
static label_def_t* label_defs = NULL;
static label_ref_t* label_refs = NULL;

static void parse_instruction(char* line);
static reg_idx_t parse_register(char* str);
static int parse_integer(char* str);
static void parse_label(char* line);
static void add_label_def(char* name, void* value);
static void* find_label_def(char* name);
static void free_label_defs();
static void add_label_ref(char* name, void* instr_addr);
static char* string_duplicate(const char* str);

void loader_load_file(char* file_name) {
  FILE* file = fopen(file_name, "r");
  assert(file != NULL);

  for (;;) {
    char line[500];
    if (fgets(line, sizeof(line), file) == NULL)
      break;

    // remove newline characters
    for (int i = strlen(line) - 1;
         i >= 0 && (line[i] == '\n' || line[i] == '\r');
         --i)
      line[i] = '\0';

    switch (line[0]) {
    case ';':
      break;                    // comment, skip line
    case ' ':
    case '\t':
      parse_instruction(line);
      break;
    default:
      parse_label(line);
      break;
    }
  }

  if (label_refs != NULL)
    fail("unresolved label(s): %s, ...", label_refs->name);
  free_label_defs();

  fclose(file);
}

static void parse_instruction(char* line) {
  char* sep = " \t";
  char* opcode_str = strtok(line, sep);
  opcode_t opcode;

  if (strcmp(opcode_str, "ADD") == 0)
    opcode = opcode_ADD;
  else if (strcmp(opcode_str, "SUB") == 0)
    opcode = opcode_SUB;
  else if (strcmp(opcode_str, "MUL") == 0)
    opcode = opcode_MUL;
  else if (strcmp(opcode_str, "DIV") == 0)
    opcode = opcode_DIV;
  else if (strcmp(opcode_str, "MOD") == 0)
    opcode = opcode_MOD;
  else if (strcmp(opcode_str, "JLT") == 0)
    opcode = opcode_JLT;
  else if (strcmp(opcode_str, "JLE") == 0)
    opcode = opcode_JLE;
  else if (strcmp(opcode_str, "JEQ") == 0)
    opcode = opcode_JEQ;
  else if (strcmp(opcode_str, "JNE") == 0)
    opcode = opcode_JNE;
  else if (strcmp(opcode_str, "JGE") == 0)
    opcode = opcode_JGE;
  else if (strcmp(opcode_str, "JGT") == 0)
    opcode = opcode_JGT;
  else if (strcmp(opcode_str, "LINT") == 0)
    opcode = opcode_LINT;
  else if (strcmp(opcode_str, "LOAD") == 0)
    opcode = opcode_LOAD;
  else if (strcmp(opcode_str, "STOR") == 0)
    opcode = opcode_STOR;
  else if (strcmp(opcode_str, "ALOC") == 0)
    opcode = opcode_ALOC;
  else if (strcmp(opcode_str, "RCHR") == 0)
    opcode = opcode_RCHR;
  else if (strcmp(opcode_str, "PCHR") == 0)
    opcode = opcode_PCHR;
  else if (strcmp(opcode_str, "HALT") == 0)
    opcode = opcode_HALT;
  else
    fail("unknown opcode %s (warning: case sensitive)", opcode_str);

  switch (opcode) {
  case opcode_ADD:
  case opcode_SUB:
  case opcode_MUL:
  case opcode_DIV:
  case opcode_MOD:
  case opcode_JLT:
  case opcode_JLE:
  case opcode_JEQ:
  case opcode_JNE:
  case opcode_JGE:
  case opcode_JGT:
  case opcode_LOAD:
  case opcode_STOR: {
    reg_idx_t ra = parse_register(strtok(NULL, sep));
    reg_idx_t rb = parse_register(strtok(NULL, sep));
    reg_idx_t rc = parse_register(strtok(NULL, sep));
    engine_emit_rrr(opcode, ra, rb, rc);
  } break;

  case opcode_ALOC: {
    reg_idx_t ra = parse_register(strtok(NULL, sep));
    reg_idx_t rb = parse_register(strtok(NULL, sep));
    engine_emit_rr(opcode, ra, rb);
  } break;

  case opcode_LINT: {
    reg_idx_t r = parse_register(strtok(NULL, sep));
    char* label_or_integer = strtok(NULL, sep);
    if (label_or_integer != NULL && isdigit(label_or_integer[0]))
      engine_emit_rc(opcode, r, parse_integer(label_or_integer));
    else {
      void* label_val = find_label_def(label_or_integer);
      if (label_val == NULL) {
        add_label_ref(label_or_integer, engine_get_next_address());
        engine_emit_rc(opcode, r, 0);
      } else
        engine_emit_rc(opcode, r, (intptr_t)label_val);
    }
  } break;

  case opcode_RCHR:
  case opcode_PCHR: {
    reg_idx_t r = parse_register(strtok(NULL, sep));
    engine_emit_r(opcode, r);
  } break;

  case opcode_HALT: {
    engine_emit(opcode);
  } break;

  default:
    fail("internal error");
  }
}

static reg_idx_t parse_register(char* str) {
  if (str[0] != 'r' && str[0] != 'R')
    fail("cannot parse register %s", str);
  return atoi(&str[1]);
}

static int parse_integer(char* str) {
  return atoi(str);
}

static void parse_label(char* line) {
  char* label;
  label = strtok(line, " :");
  if (label == NULL)
    fail("cannot parse label %s", line);

  add_label_def(label, engine_get_next_address());
}

static void add_label_def(char* name, void* value) {
  // add label definition to list
  label_def_t* ldef = malloc(sizeof(label_def_t));
  assert(ldef != NULL);
  ldef->name = string_duplicate(name);
  ldef->addr = value;
  ldef->next = label_defs;
  label_defs = ldef;

  // patch all instructions that refer to the newly-defined label
  label_ref_t* lref = label_refs;
  label_ref_t** prev_ptr = &label_refs;
  while (lref != NULL) {
    label_ref_t* node_to_free = NULL;
    if (strcmp(lref->name, name) == 0) {
      engine_patch(lref->instr_addr, value);
      node_to_free = lref;
      *prev_ptr = lref->next;
    } else
      prev_ptr = &lref->next;
    lref = lref->next;
    if (node_to_free != NULL) {
      free(node_to_free->name);
      free(node_to_free);
    }
  }
}

static void* find_label_def(char* name) {
  label_def_t* ldef;
  for (ldef = label_defs;
       ldef != NULL && strcmp(ldef->name, name) != 0;
       ldef = ldef->next)
    ;
  return (ldef == NULL) ? NULL : ldef->addr;
}

static void free_label_defs() {
  label_def_t* ldef = label_defs;
  while (ldef != NULL) {
    label_def_t* next = ldef->next;
    free(ldef->name);
    free(ldef);
    ldef = next;
  }
}

static void add_label_ref(char* name, void* instr_addr) {
  label_ref_t* lref = malloc(sizeof(label_ref_t));
  lref->name = string_duplicate(name);
  lref->instr_addr = instr_addr;
  lref->next = label_refs;
  label_refs = lref;
}

static char* string_duplicate(const char* str) {
  char* dup = malloc(strlen(str) + 1);
  strcpy(dup, str);
  return dup;
}
